@@ -1,3 +1,4 @@
+#include <stdio.h>
 #include <arpa/inet.h>
 #include <netinet/in.h>
 #include <stdint.h>
@@ -16,7 +17,6 @@ using namespace std;
 
 Client::Client(int sock)
 	: sock(sock),
-	  connect_time(time(NULL)),
 	  state(Client::READING_REQUEST),
 	  stream(NULL),
 	  header_or_error_bytes_sent(0),
@@ -27,6 +27,11 @@ Client::Client(int sock)
 {
 	request.reserve(1024);
 
+	if (clock_gettime(CLOCK_MONOTONIC_COARSE, &connect_time) == -1) {
+		log_perror("clock_gettime(CLOCK_MONOTONIC_COARSE)");
+		return;
+	}
+
 	// Find the remote address, and convert it to ASCII.
 	sockaddr_in6 addr;
 	socklen_t addr_len = sizeof(addr);
@@ -59,7 +64,8 @@ Client::Client(int sock)
 Client::Client(const ClientProto &serialized, Stream *stream)
 	: sock(serialized.sock()),
 	  remote_addr(serialized.remote_addr()),
-	  connect_time(serialized.connect_time()),
+	  referer(serialized.referer()),
+	  user_agent(serialized.user_agent()),
 	  state(State(serialized.state())),
 	  request(serialized.request()),
 	  url(serialized.url()),
@@ -78,6 +84,8 @@ Client::Client(const ClientProto &serialized, Stream *stream)
 			}
 		}
 	}
+	connect_time.tv_sec = serialized.connect_time_sec();
+	connect_time.tv_nsec = serialized.connect_time_nsec();
 }
 
 ClientProto Client::serialize() const
@@ -85,7 +93,10 @@ ClientProto Client::serialize() const
 	ClientProto serialized;
 	serialized.set_sock(sock);
 	serialized.set_remote_addr(remote_addr);
-	serialized.set_connect_time(connect_time);
+	serialized.set_referer(referer);
+	serialized.set_user_agent(user_agent);
+	serialized.set_connect_time_sec(connect_time.tv_sec);
+	serialized.set_connect_time_nsec(connect_time.tv_nsec);
 	serialized.set_state(state);
 	serialized.set_request(request);
 	serialized.set_url(url);
@@ -97,6 +108,24 @@ ClientProto Client::serialize() const
 	serialized.set_num_loss_events(num_loss_events);
 	return serialized;
 }
+
+namespace {
+
+string escape_string(const string &str) {
+	string ret;
+	for (size_t i = 0; i < str.size(); ++i) {
+		char buf[16];
+		if (isprint(str[i]) && str[i] >= 32 && str[i] != '"' && str[i] != '\\') {
+			ret.push_back(str[i]);
+		} else {
+			snprintf(buf, sizeof(buf), "\\x%02x", (unsigned char)str[i]);
+			ret += buf;
+		}
+	}
+	return ret;
+}
+
+} // namespace
 	
 ClientStats Client::get_stats() const
 {
@@ -108,6 +137,8 @@ ClientStats Client::get_stats() const
 	}
 	stats.sock = sock;
 	stats.remote_addr = remote_addr;
+	stats.referer = escape_string(referer);
+	stats.user_agent = escape_string(user_agent);
 	stats.connect_time = connect_time;
 	stats.bytes_sent = bytes_sent;
 	stats.bytes_lost = bytes_lost;
