@@ -11,11 +11,14 @@ class ClientProto;
 struct Stream;
 
 // Digested statistics for writing to logs etc.
+// Note that "referer" and "user_agent" here are already escaped for scary characters.
 struct ClientStats {
 	std::string url;
 	int sock;
 	std::string remote_addr;
-	time_t connect_time;
+	std::string referer;
+	std::string user_agent;
+	timespec connect_time;
 	size_t bytes_sent;
 	size_t bytes_lost;
 	size_t num_loss_events;
@@ -33,11 +36,15 @@ struct Client {
 	// The file descriptor associated with this socket.
 	int sock;
 
+	// When the client connected, in terms of CLOCK_MONOTONIC_COARSE.
+	timespec connect_time;
+
 	// Some information only used for logging.
 	std::string remote_addr;
-	time_t connect_time;
+	std::string referer;
+	std::string user_agent;
 
-	enum State { READING_REQUEST, SENDING_HEADER, SENDING_DATA, SENDING_ERROR, WAITING_FOR_KEYFRAME };
+	enum State { READING_REQUEST, SENDING_HEADER, SENDING_DATA, SENDING_ERROR, WAITING_FOR_KEYFRAME, PREBUFFERING };
 	State state;
 
 	// The HTTP request, as sent by the client. If we are in READING_REQUEST,
@@ -62,8 +69,8 @@ struct Client {
 	// -1 means we want to send from the end of the backlog (the normal case),
 	// although only at a keyframe.
 	// -2 means we want to send from the _beginning_ of the backlog.
-	// Once we go into WAITING_FOR_KEYFRAME or SENDING_DATA, these negative
-	// values will be translated to real numbers.
+	// Once we go into WAITING_FOR_KEYFRAME, PREBUFFERING or SENDING_DATA,
+	// these negative values will be translated to real numbers.
 	size_t stream_pos;
 
 	// Number of bytes we've sent of data. Only relevant for SENDING_DATA.
