@@ -10,6 +10,7 @@
 #include <sys/time.h>
 #include <sys/wait.h>
 #include <unistd.h>
+#include <algorithm>
 #include <map>
 #include <set>
 #include <string>
@@ -37,6 +38,18 @@ ServerPool *servers = NULL;
 volatile bool hupped = false;
 volatile bool stopped = false;
 
+namespace {
+
+struct OrderByConnectionTime {
+	bool operator() (const ClientProto &a, const ClientProto &b) const {
+		if (a.connect_time_sec() != b.connect_time_sec())
+			return a.connect_time_sec() < b.connect_time_sec();
+		return a.connect_time_nsec() < b.connect_time_nsec();
+	}
+};
+
+}  // namespace
+
 struct InputWithRefcount {
 	Input *input;
 	int refcount;
@@ -54,14 +67,14 @@ void do_nothing(int signum)
 {
 }
 
-CubemapStateProto collect_state(const timeval &serialize_start,
+CubemapStateProto collect_state(const timespec &serialize_start,
                                 const vector<Acceptor *> acceptors,
                                 const multimap<string, InputWithRefcount> inputs,
                                 ServerPool *servers)
 {
 	CubemapStateProto state = servers->serialize();  // Fills streams() and clients().
 	state.set_serialize_start_sec(serialize_start.tv_sec);
-	state.set_serialize_start_usec(serialize_start.tv_usec);
+	state.set_serialize_start_usec(serialize_start.tv_nsec / 1000);
 	
 	for (size_t i = 0; i < acceptors.size(); ++i) {
 		state.add_acceptors()->MergeFrom(acceptors[i]->serialize());
@@ -166,11 +179,13 @@ void create_streams(const Config &config,
 		if (deserialized_urls.count(stream_config.url) == 0) {
 			stream_index = servers->add_stream(stream_config.url,
 			                                   stream_config.backlog_size,
+			                                   stream_config.prebuffering_bytes,
 			                                   Stream::Encoding(stream_config.encoding));
 		} else {
 			stream_index = servers->lookup_stream_by_url(stream_config.url);
 			assert(stream_index != -1);
 			servers->set_backlog_size(stream_index, stream_config.backlog_size);
+			servers->set_prebuffering_bytes(stream_index, stream_config.prebuffering_bytes);
 			servers->set_encoding(stream_index,
 			                      Stream::Encoding(stream_config.encoding));
 		}
@@ -375,7 +390,7 @@ start:
 	find_deleted_streams(config, &deleted_urls);
 
 	CubemapStateProto loaded_state;
-	struct timeval serialize_start;
+	timespec serialize_start;
 	set<string> deserialized_urls;
 	map<sockaddr_in6, Acceptor *, Sockaddr6Compare> deserialized_acceptors;
 	multimap<string, InputWithRefcount> inputs;  // multimap due to older versions without deduplication.
@@ -391,7 +406,7 @@ start:
 		}
 
 		serialize_start.tv_sec = loaded_state.serialize_start_sec();
-		serialize_start.tv_usec = loaded_state.serialize_start_usec();
+		serialize_start.tv_nsec = loaded_state.serialize_start_usec() * 1000ull;
 
 		// Deserialize the streams.
 		map<string, string> stream_headers_for_url;  // See below.
@@ -443,9 +458,33 @@ start:
 	// Find all streams in the configuration file, create them, and connect to the inputs.
 	create_streams(config, deserialized_urls, &inputs);
 	vector<Acceptor *> acceptors = create_acceptors(config, &deserialized_acceptors);
+
+	// Convert old-style timestamps to new-style timestamps for all clients;
+	// this simplifies the sort below.
+	{
+		timespec now_monotonic;
+		if (clock_gettime(CLOCK_MONOTONIC_COARSE, &now_monotonic) == -1) {
+			log(ERROR, "clock_gettime(CLOCK_MONOTONIC_COARSE) failed.");
+			exit(1);
+		}
+		long delta_sec = now_monotonic.tv_sec - time(NULL);
+
+		for (int i = 0; i < loaded_state.clients_size(); ++i) {
+			ClientProto* client = loaded_state.mutable_clients(i);
+			if (client->has_connect_time_old()) {
+				client->set_connect_time_sec(client->connect_time_old() + delta_sec);
+				client->set_connect_time_nsec(now_monotonic.tv_nsec);
+				client->clear_connect_time_old();
+			}
+		}
+	}
 	
 	// Put back the existing clients. It doesn't matter which server we
-	// allocate them to, so just do round-robin.
+	// allocate them to, so just do round-robin. However, we need to sort them
+	// by connection time first, since add_client_serialized() expects that.
+	sort(loaded_state.mutable_clients()->begin(),
+	     loaded_state.mutable_clients()->end(),
+	     OrderByConnectionTime());
 	for (int i = 0; i < loaded_state.clients_size(); ++i) {
 		if (deleted_urls.count(loaded_state.clients(i).url()) != 0) {
 			safe_close(loaded_state.clients(i).sock());
@@ -490,14 +529,15 @@ start:
 		input_stats_thread->run();
 	}
 
-	struct timeval server_start;
-	gettimeofday(&server_start, NULL);
+	timespec server_start;
+	int err = clock_gettime(CLOCK_MONOTONIC, &server_start);
+	assert(err != -1);
 	if (state_fd != -1) {
 		// Measure time from we started deserializing (below) to now, when basically everything
 		// is up and running. This is, in other words, a conservative estimate of how long our
 		// “glitch” period was, not counting of course reconnects if the configuration changed.
 		double glitch_time = server_start.tv_sec - serialize_start.tv_sec +
-			1e-6 * (server_start.tv_usec - serialize_start.tv_usec);
+			1e-9 * (server_start.tv_nsec - serialize_start.tv_nsec);
 		log(INFO, "Re-exec happened in approx. %.0f ms.", glitch_time * 1000.0);
 	}
 
@@ -506,7 +546,8 @@ start:
 	}
 
 	// OK, we've been HUPed. Time to shut down everything, serialize, and re-exec.
-	gettimeofday(&serialize_start, NULL);
+	err = clock_gettime(CLOCK_MONOTONIC, &serialize_start);
+	assert(err != -1);
 
 	if (input_stats_thread != NULL) {
 		input_stats_thread->stop();
