@@ -8,6 +8,7 @@
 #include <sys/types.h>
 #include <time.h>
 #include <map>
+#include <queue>
 #include <string>
 #include <vector>
 
@@ -21,6 +22,7 @@ struct Stream;
 #define EPOLL_MAX_EVENTS 8192
 #define EPOLL_TIMEOUT_MS 20
 #define MAX_CLIENT_REQUEST 16384
+#define REQUEST_READ_TIMEOUT_SEC 60
 
 class CubemapStateProto;
 class StreamProto;
@@ -53,10 +55,11 @@ public:
 	// at the same time).
 	CubemapStateProto serialize();
 	void add_client_from_serialized(const ClientProto &client);
-	int add_stream(const std::string &url, size_t bytes_received, Stream::Encoding encoding);
+	int add_stream(const std::string &url, size_t bytes_received, size_t prebuffering_bytes, Stream::Encoding encoding);
 	int add_stream_from_serialized(const StreamProto &stream, int data_fd);
 	int lookup_stream_by_url(const std::string &url) const;
 	void set_backlog_size(int stream_index, size_t new_size);
+	void set_prebuffering_bytes(int stream_index, size_t new_amount);
 	void set_encoding(int stream_index, Stream::Encoding encoding);
 
 private:
@@ -89,6 +92,19 @@ private:
 	// Map from file descriptor to client.
 	std::map<int, Client> clients;
 
+	// A list of all clients, ordered by the time they connected (first element),
+	// and their file descriptor (second element). It is ordered by connection time
+	// (and thus also by read timeout time) so that we can read clients from the
+	// start and stop processing once we get to one that isn't ready to be
+	// timed out yet (which means we only have to look at each client exactly once,
+	// save for the first element of the queue, which is always checked).
+	//
+	// Note that when we delete a client, we don't update this queue.
+	// This means that when reading it, we need to check if the client it
+	// describes is still exists (ie., that the fd still exists, and that
+	// the timespec matches).
+	std::queue<std::pair<timespec, int> > clients_ordered_by_connect_time;
+
 	// Used for epoll implementation (obviously).
 	int epoll_fd;
 	epoll_event events[EPOLL_MAX_EVENTS];
