@@ -73,14 +73,14 @@ int ServerPool::lookup_stream_by_url(const std::string &url) const
 	return servers[0].lookup_stream_by_url(url);
 }
 
-int ServerPool::add_stream(const string &url, size_t backlog_size, Stream::Encoding encoding)
+int ServerPool::add_stream(const string &url, size_t backlog_size, size_t prebuffering_bytes, Stream::Encoding encoding)
 {
 	// Adding more HTTP streams after UDP streams would cause the UDP stream
 	// indices to move around, which is obviously not good.
 	assert(udp_streams.empty());
 
 	for (int i = 0; i < num_servers; ++i) {
-		int stream_index = servers[i].add_stream(url, backlog_size, encoding);
+		int stream_index = servers[i].add_stream(url, backlog_size, prebuffering_bytes, encoding);
 		assert(stream_index == num_http_streams);
 	}
 	return num_http_streams++;
@@ -201,6 +201,13 @@ void ServerPool::set_backlog_size(int stream_index, size_t new_size)
 	}	
 }
 
+void ServerPool::set_prebuffering_bytes(int stream_index, size_t new_amount)
+{
+	for (int i = 0; i < num_servers; ++i) {
+		servers[i].set_prebuffering_bytes(stream_index, new_amount);
+	}
+}
+
 void ServerPool::set_encoding(int stream_index, Stream::Encoding encoding)
 {
 	for (int i = 0; i < num_servers; ++i) {
